package be.iminds.ilabt.jfed.restapi.resource;

import be.iminds.ilabt.jfed.lowlevel.AnyCredential;
import be.iminds.ilabt.jfed.lowlevel.CredentialException;
import be.iminds.ilabt.jfed.lowlevel.SimpleGeniUserWithSpeaksFor;
import be.iminds.ilabt.jfed.lowlevel.userloginmodel.InvalidLoginException;
import be.iminds.ilabt.jfed.restapi.JFedRestApiConfiguration;
import be.iminds.ilabt.jfed.restapi.dao.SpeaksForCredentialDao;
import be.iminds.ilabt.jfed.restapi.jfedclient.JFedExperimenterApi;
import be.iminds.ilabt.jfed.restapi.jfedclient.JFedExperimenterApiImpl;
import be.iminds.ilabt.jfed.restapi.json.SpeaksForCredential;
import be.iminds.ilabt.jfed.restapi.json.User;
import be.iminds.ilabt.jfed.restapi.json.UserBuilder;
import be.iminds.ilabt.jfed.util.GeniUrn;
import com.codahale.metrics.annotation.Timed;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiOperation;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.servlet.http.HttpServletRequest;
import javax.validation.constraints.NotNull;
import javax.ws.rs.*;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.UriInfo;
import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Collections;
import java.util.List;

@Path("/user")
@Api(value = "user")
@Produces(MediaType.APPLICATION_JSON)
public class UserResource {
    private static final Logger LOG = LoggerFactory.getLogger(UserResource.class);

    @Context
    UriInfo uriInfo;

    private final JFedRestApiConfiguration configuration;
    private final SpeaksForCredentialDao speaksForCredentialDao;

    public UserResource(JFedRestApiConfiguration configuration, SpeaksForCredentialDao speaksForCredentialDao) {
        this.configuration = configuration;
        this.speaksForCredentialDao = speaksForCredentialDao;
    }

    @GET
    @Path("/")
    @ApiOperation(value = "Request info on yourself")
    @Timed
    public User getUserInfo(@Context HttpServletRequest request) {
        GeniUrn authenticatedUser = configuration.getAuthenticatedUser(request);
        if (authenticatedUser == null) {
            throw new WebApplicationException("The SSL connection did not correctly authenticate you as a trusted user. Did you use SSL client authentication?",
                    Response.Status.FORBIDDEN);
        }

        return getUserInfo(authenticatedUser.getEncodedTopLevelAuthority(), authenticatedUser.getEncodedResourceName(), request);
    }

    @GET
    @Path("{tld}/{name}")
    @Timed
    public User getUserInfo(@NotNull @PathParam("tld") String tld, @NotNull @PathParam("name") String name, @Context HttpServletRequest request) {
        GeniUrn authenticatedUser = configuration.getAuthenticatedUser(request);
        if (authenticatedUser == null) {
            throw new WebApplicationException("The SSL connection did not correctly authenticate you as a trusted user. Did you use SSL client authentication?",
                    Response.Status.FORBIDDEN);
        }

        //TODO: add debugdata
//        //remove any leftover experimentId from MDC
//        MDC.remove("experimentId");

        GeniUrn requestedUser = GeniUrn.createGeniUrnFromEncodedParts(tld, "user", name);

        if (!authenticatedUser.equals(requestedUser) && !configuration.hasAccess(JFedRestApiConfiguration.Access.ADMIN, request)) {
            LOG.warn("FORBIDDING ACCESS: You are authenticated as user " + authenticatedUser + ". You do not have access to info about " + requestedUser);
            throw new WebApplicationException("You are authenticated as user " + authenticatedUser + ". You do not have access to info about " + requestedUser,
                    Response.Status.FORBIDDEN);
        }

        SpeaksForCredential speaksForCredentialInfo =
                speaksForCredentialDao.find(configuration.getUser().getUserUrn(), requestedUser);

        if (speaksForCredentialInfo == null) {
            LOG.warn("NO SPEAKSFOR REGISTERED: You need to register a speaksfor credential to allow " +
                    configuration.getUser().getUserUrn() + " to speak for " + authenticatedUser);
            throw new WebApplicationException("NO SPEAKSFOR REGISTERED: You need to register a speaksfor credential to allow " +
                    configuration.getUser().getUserUrn() + " to speak for " + authenticatedUser,
                    Response.Status.BAD_REQUEST);
        }

        AnyCredential speaksForCredential = null;
        try {
            speaksForCredential = AnyCredential.createAutoDetect("speaksfor credential from DB",
                    speaksForCredentialInfo.getCredentialXml());
        } catch (CredentialException e) {
            throw new WebApplicationException("An invalid speaksfor credential is stored for " +
                    configuration.getUser().getUserUrn() + " speaking for " + authenticatedUser, e,
                    Response.Status.INTERNAL_SERVER_ERROR);
        }

        assert speaksForCredential != null;
        JFedExperimenterApi jFedExperimenterApi = null;
        try {
            jFedExperimenterApi = new JFedExperimenterApiImpl(
                    new File(configuration.getSfaUserLoginPemFilename()),
                    null,
                    Collections.singletonList(speaksForCredential),
                    requestedUser
            );
        } catch (IOException e) {
            throw new WebApplicationException("Error setting up JFedExperimenterApi: " + e.getMessage(), e,
                    Response.Status.INTERNAL_SERVER_ERROR);
        } catch (SimpleGeniUserWithSpeaksFor.InvalidSpeaksForException e) {
            throw new WebApplicationException("An invalid speaksfor credential is stored for " +
                    configuration.getUser().getUserUrn() + " speaking for " + authenticatedUser, e,
                    Response.Status.INTERNAL_SERVER_ERROR);
        } catch (InvalidLoginException e) {
            throw new WebApplicationException("Server configuration error: speaking user login PEM file not valid", e,
                    Response.Status.INTERNAL_SERVER_ERROR);
        }

        List<GeniUrn> projects = jFedExperimenterApi.getProjectsForUser(requestedUser);
        List<GeniUrn> slices = jFedExperimenterApi.getSlicesForUser(requestedUser);

        UserBuilder res = new UserBuilder();
        res.setId(requestedUser.getResourceName());
        res.setUrn(requestedUser);
        res.setAuthorityTld(requestedUser.getTopLevelAuthority_withoutSubAuth());
        for (GeniUrn projectUrn : projects) {
            res.addProject(projectUrn);
        }
        for (GeniUrn sliceUrn : slices) {
            res.addSlice(sliceUrn);
        }

        try {
            addUri(res);
        } catch (Exception e) {
            LOG.error("Failed to set URI", e);
        }

        return res.create();
    }


    public UserBuilder addUri(UserBuilder userBuilder) {
        String base =
                configuration == null || configuration.getBaseUrl() == null || configuration.getBaseUrl().trim().isEmpty()
                        ?
                        null : configuration.getBaseUrl();
        if (base != null && !base.endsWith("/")) base += "/";

        GeniUrn userUrn = userBuilder.getUrn();
        assert userUrn != null;

        URI uri = null;
        try {
            uri =
                    new URI(base + "user/" + userUrn.getEncodedTopLevelAuthority_withoutSubAuth() + "/" + userUrn.getEncodedResourceName());
            userBuilder.setUri(uri);
        } catch (URISyntaxException e) {
            LOG.error("Error when adding URI", e);
            //throw new WebApplicationException("Error when adding URI", Response.Status.INTERNAL_SERVER_ERROR);
            //ignore
        }

        return userBuilder;
    }
}
